require 'active_record'

require 'ar-extensions'
require 'ar-extensions/adapters/mysql'
require 'ar-extensions/import/mysql'

class Summarizer
  include BaselineUtil

  def self.get_some_status_data_quickly(id)
    sql = <<SQL
SELECT
    resource_statuses.resource_type,
    resource_statuses.title,
    resource_events.property,
    resource_events.previous_value
FROM
    resource_statuses
INNER JOIN
    resource_events ON resource_events.resource_status_id = resource_statuses.id
WHERE
    resource_statuses.report_id = #{id}
    AND resource_statuses.failed = 0
SQL

    rset = ActiveRecord::Base.connection.raw_connection.query(sql)

    result = Set.new
    rset.each {|row| result << row }
    rset.free

    return result
  end

  def self.get_baseline_data_quickly(id)
    sql = <<SQL
SELECT
    baseline_resources.resource_type,
    baseline_resources.title,
    baseline_properties.name,
    baseline_properties.value
FROM
    baseline_resources
INNER JOIN
    baseline_properties ON baseline_resources.baseline_id = #{id}
    AND baseline_properties.baseline_resource_id = baseline_resources.id
SQL

    rset = ActiveRecord::Base.connection.raw_connection.query(sql)

    result = Set.new
    rset.each {|row| result << row}
    rset.free

    return result
  end

  def self.get_some_resource_status_ids_quickly(report)
    sql = <<SQL
SELECT
    id, resource_type, title
FROM
    resource_statuses
WHERE
    report_id = #{report}
SQL

    rset = ActiveRecord::Base.connection.raw_connection.query(sql)

    result = {}
    rset.each {|row| result[[row[1], row[2]]] = row[0] }
    rset.free

    return result
  end

  def self.get_some_baseline_resources_quickly(baseline)
    sql = <<SQL
SELECT
    id, resource_type, title
FROM
    baseline_resources
WHERE
    baseline_id = #{baseline}
SQL

    rset = ActiveRecord::Base.connection.raw_connection.query(sql)

    result = {}
    rset.each {|row| result[[row[1], row[2]]] = row[0] }
    rset.free

    return result
  end

  def self.calculate_some_baseline_property_hashes_quickly(baseline_resource_ids)
    return {} unless baseline_resource_ids.present?

    sql = <<SQL
SELECT
  baseline_resource_id, name, value
FROM
  baseline_properties
WHERE
  baseline_resource_id IN (#{baseline_resource_ids.join(', ')})
ORDER BY
  name ASC
SQL

    calculate_resource_digests(sql)
  end

  def self.calculate_some_resource_event_hashes_quickly(resource_status_ids)
    return {} unless resource_status_ids.present?

    # This quickly calculates the MD5 hash of the outcome of a particular
    # change ID for later use.  This uses the *ordered* set of property names
    # and values to calculate the checksum.
    #
    # That means that the "ORDER BY" in the SQL is important in providing that
    # assurance to the streaming calculation that follows. --daniel 2011-06-21
    sql = <<SQL
SELECT
  resource_status_id, property, previous_value
FROM
  resource_events
WHERE
  resource_status_id IN (#{resource_status_ids.join(', ')})
ORDER BY
  property ASC
SQL

    calculate_resource_digests(sql)
  end

  def self.calculate_resource_digests(sql)
    rset = ActiveRecord::Base.connection.raw_connection.query(sql)

    result = {}
    rset.each do |row|
      result[row[0]] ||= Digest::MD5.new
      result[row[0]] << row[1..-1].join('')
    end
    rset.free

    # Get the digest, base64 encode, and chomp off the newline that Ruby wants
    # to throw on the end.  Since we only mutate the values, we shouldn't
    # perturb our walk through the hash. --daniel 2011-06-21
    result.each {|k,v| result[k] = [v.digest].pack('m').chomp }

    return result
  end

  def self.compare_group_against_node(group, reference_node_id)
    reference_node_baseline = NodeBaseline.self_comparisons.newest_first.find_by_node_id(reference_node_id)
    reference_baseline = reference_node_baseline.baseline

    if custom_report = CustomReport.find_by_group_id(group)
      return true if custom_report.node_baseline_id == reference_node_baseline.id
      custom_report.delete
    end

    custom_report = CustomReport.create!(
      :group_id => group.id,
      :node_baseline_id => reference_node_baseline.id
    )

    group.nodes.each do |node|
      if node_baseline = NodeBaseline.find_by_node_id_and_date_and_reference_node_id(node.id, reference_node_baseline.date, reference_node_id)
        custom_report.node_baselines << node_baseline
        next
      end

      node_baseline = NodeBaseline.create!(
        :node_id => node.id,
        :reference_node_id => reference_node_id,
        :date => reference_node_baseline.date,
        :baseline_id => reference_baseline.id
      )

      if report_id = node.last_inspect_report_id
        delay.build_baseline_summary(report_id, node_baseline, custom_report.id)
      end
    end
    return false
  end

  def self.build_baseline_summary(report_id, node_baseline, custom_report_id = nil)
    baseline_id = node_baseline.baseline_id

    report   = get_some_status_data_quickly(report_id)
    baseline = get_baseline_data_quickly(baseline_id)
    delta    = (report ^ baseline)

    BaselineDifference.transaction do
      node_baseline.update_attributes!(:diffed => true)
      if custom_report_id && custom_report = CustomReport.find(custom_report_id)
        custom_report.node_baselines << node_baseline
        custom_report.save!
      end

      unless delta.empty?
        report_statuses           = get_some_resource_status_ids_quickly(report_id)
        baseline_statuses         = get_some_baseline_resources_quickly(baseline_id)
        report_status_digests     = calculate_some_resource_event_hashes_quickly(report_statuses.values)
        baseline_resource_digests = calculate_some_baseline_property_hashes_quickly(baseline_statuses.values)

        rows = delta.map do |diff|
          key = [diff[0], diff[1]]

          report_resource          = report_statuses[key]
          baseline_resource        = baseline_statuses[key]
          resource_name            = "#{diff[0]}[#{diff[1]}]"
          report_resource_digest   = report_status_digests[report_resource]
          baseline_resource_digest = baseline_resource_digests[baseline_resource]

          [node_baseline.id, report_resource, baseline_resource, resource_name,
           report_resource_digest, baseline_resource_digest]
        end.uniq

        BaselineDifference.import(
          [:node_baseline_id, :report_resource_id, :baseline_resource_id,
           :resource_name, :resource_status_digest, :baseline_resource_digest],
          rows
        )
      end
    end

    return nil                  # no meaningful result available!
  end

  def self.queue_diff_summarization
    nodes_with_baselines = NodeBaseline.self_comparisons.
      find_all_by_diffed(false).group_by(&:node_id)

    nodes_with_baselines.each do |node_id, node_baselines|
      node_baselines.each do |nb|
        time = BaselineUtil.day_beginning(nb.date)..BaselineUtil.day_ending(nb.date)
        report = Report.find(:first, :conditions => {
                               :kind    => 'inspect',
                               :time    => time,
                               :node_id => node_id},
                             :order => 'time DESC',
                             :select => 'id, node_id')

        # If we have not received the report yet, we just carry on.  When the
        # report finally comes in, if even, we will finally do the diff then.
        #
        # Eventually we will need a way to delete node_baseline records, or
        # mark them as not still pending, but for now...
        if report then
          self.delay.build_baseline_summary(report.id, nb)
        end
        yield if block_given?
      end
    end
  end
end
